"""
Created on Mar 26, 2013

@author: mmikofski
"""


class PVexception(Exception):
    """
    Base exception class for pvmismatch.
    """
    pass


class PVparallel_calcError(PVexception):
    def __init__(self, name):  # IGNORE:W0231
        self.name = name

    def __str__(self):
        errmsg = ('Parallel function from module, %s, must be on the main' +
                  ' thread.')
        return errmsg % self.name



# -*- coding: utf-8 -*-
"""Vectorized pvmodule."""

import numpy as np
from .pvmismatch import pvconstants
from .utils import calcMPP_IscVocFFBPD
from .circuit_comb import calcSeries, calcParallel, combine_parallel_circuits, parse_diode_config
from .circuit_comb import calcSeries_with_bypass, calcParallel_with_bypass
from .circuit_comb import DEFAULT_BYPASS, MODULE_BYPASS, CUSTOM_SUBSTR_BYPASS
# ------------------------------------------------------------------------------
# CALCULATE MODULE IV-PV CURVES------------------------------------------------


def calcMods(cell_pos, maxmod, cell_index_map, Ee_mod, Ee_cell, u_cell_type, cell_type, cell_data,
             outer_circuit, run_bpact=True):
    """
    Generate all module IV curves and store results in a dictionary.

    Parameters
    ----------
    cell_pos : dict
        cell position pattern from pvmismatch package.
    maxmod : pvmodule object
        pvmodule class from pvmismatch package.
    cell_index_map : numpy.ndarray
        2-D array specifying the physical cell positions in the module.
    Ee_mod : numpy.ndarray
        3-D array containing the Irradiance at the cell level for all modules.
    Ee_cell : numpy.ndarray
        1-D array containing irradiances in suns.
    u_cell_type : list
        List of cell types at each irradiance setting.
    cell_type : numpy.ndarray
        2-D array of cell types for each cell in each module.
    cell_data : dict
        Dictionary containing cell IV curves.
    outer_circuit : str
        series or parallel.
    run_bpact : bool, optional
        Flag to run bypass diode activation logic. The default is True.

    Returns
    -------
    mod_data : dict
        Dictionary containing module IV curves.

    """
    Vbypass = maxmod.Vbypass
    I_mod_curves = []
    V_mod_curves = []
    P_mod_curves = []
    Isubstr_curves = []
    Vsubstr_curves = []
    Isubstr_pre_bypass_curves = []
    Vsubstr_pre_bypass_curves = []
    mean_Iscs = []
    bypassed_mod_arr = []
    for idx_mod in range(Ee_mod.shape[0]):
        # 1 Module
        cell_ids = cell_index_map.flatten()
        idx_sort = np.argsort(cell_ids)
        Ee_mod1 = Ee_mod[idx_mod].flatten()[idx_sort]
        cell_type1 = cell_type.flatten()[idx_sort]
        # Extract cell IV curves
        # mod_in_cell = np.where(np.in1d(Ee_cell, Ee_mod1) & np.in1d(u_cell_type, cell_type1))[0]
        mod_in_cell = np.where(np.in1d(Ee_cell+u_cell_type, Ee_mod1+cell_type1))[0]
        Icell_red = cell_data['Icell'][mod_in_cell, :]
        Vcell_red = cell_data['Vcell'][mod_in_cell, :]
        Vrbd_red = cell_data['VRBD'][mod_in_cell]
        Voc_red = cell_data['Voc'][mod_in_cell]
        Isc_red = cell_data['Isc'][mod_in_cell]
        u, inverse, counts = np.unique(Ee_mod1+cell_type1, return_inverse=True,
                                       return_counts=True)
        # Expand for Mod curves
        Icell = Icell_red[inverse, :]
        Vcell = Vcell_red[inverse, :]
        VRBD = Vrbd_red[inverse]
        Voc = Voc_red[inverse]
        Isc = Isc_red[inverse]
        NPT_dict = cell_data['NPT']
        # Run Module Circuit model
        calM_op = calcMod(Icell,
                          Vcell, VRBD, Voc,
                          Isc, cell_pos,
                          Vbypass,
                          NPT_dict, outer=outer_circuit,
                          run_bpact=run_bpact)
        Imod, Vmod, Pmod, Isubstr, Vsubstr, mean_Isc, Isubstr_pre_bypass, Vsubstr_pre_bypass, bypassed_mod = calM_op
        I_mod_curves.append(np.reshape(Imod, (1, len(Imod))))
        V_mod_curves.append(np.reshape(Vmod, (1, len(Vmod))))
        P_mod_curves.append(np.reshape(Pmod, (1, len(Pmod))))
        Isubstr_curves.append(np.reshape(
            Isubstr, (1, Isubstr.shape[0], Isubstr.shape[1])))
        Vsubstr_curves.append(np.reshape(
            Vsubstr, (1, Vsubstr.shape[0], Vsubstr.shape[1])))
        Isubstr_pre_bypass_curves.append(np.reshape(
            Isubstr_pre_bypass, (1, Isubstr_pre_bypass.shape[0], Isubstr_pre_bypass.shape[1])))
        Vsubstr_pre_bypass_curves.append(np.reshape(
            Vsubstr_pre_bypass, (1, Vsubstr_pre_bypass.shape[0], Vsubstr_pre_bypass.shape[1])))
        mean_Iscs.append(mean_Isc)
        if run_bpact:
            bypassed_mod_arr.append(np.reshape(
                bypassed_mod, (1, bypassed_mod.shape[0], bypassed_mod.shape[1])))
        else:
            bypassed_mod_arr.append(bypassed_mod)
    I_mod_curves = np.concatenate(I_mod_curves, axis=0)
    V_mod_curves = np.concatenate(V_mod_curves, axis=0)
    P_mod_curves = np.concatenate(P_mod_curves, axis=0)
    Isubstr_curves = np.concatenate(Isubstr_curves, axis=0)
    Vsubstr_curves = np.concatenate(Vsubstr_curves, axis=0)
    Isubstr_pre_bypass_curves = np.concatenate(Isubstr_pre_bypass_curves, axis=0)
    Vsubstr_pre_bypass_curves = np.concatenate(Vsubstr_pre_bypass_curves, axis=0)
    mean_Iscs = np.array(mean_Iscs)
    if run_bpact:
        bypassed_mod_arr = np.concatenate(bypassed_mod_arr, axis=0)
    else:
        bypassed_mod_arr = np.array(bypassed_mod_arr)

    Imp, Vmp, Pmp, Isc, Voc, FF, BpDmp, num_bpd_active = calcMPP_IscVocFFBPD(
        I_mod_curves, V_mod_curves, P_mod_curves, bypassed_mod_arr, run_bpact=run_bpact)
    # print('Time elapsed to generate IV Curve of all unique modules: ' + str(time.time() - t0) + ' s')

    # Store results in a dict
    mod_data = dict()
    mod_data['Imod'] = I_mod_curves
    mod_data['Vmod'] = V_mod_curves
    mod_data['Pmod'] = P_mod_curves
    mod_data['Isubstr'] = Isubstr_curves
    mod_data['Vsubstr'] = Vsubstr_curves
    mod_data['Isubstr_pre_bypass'] = Isubstr_pre_bypass_curves
    mod_data['Vsubstr_pre_bypass'] = Vsubstr_pre_bypass_curves
    mod_data['Bypassed_substr'] = bypassed_mod_arr
    mod_data['mean_Isc'] = np.reshape(mean_Iscs, (len(mean_Iscs), 1))
    mod_data['Imp'] = Imp
    mod_data['Vmp'] = Vmp
    mod_data['Pmp'] = Pmp
    mod_data['Isc'] = Isc
    mod_data['Voc'] = Voc
    mod_data['FF'] = FF
    mod_data['BPDiode_Active_MPP'] = BpDmp
    mod_data['num_bpd_active'] = num_bpd_active

    return mod_data


def calcMod(Icell, Vcell, VRBD, Voc, Isc, cell_pos, Vbypass, NPT_dict,
            outer='series', run_bpact=True):
    """
    Calculate module I-V curves.

    Returns module currents [A], voltages [V] and powers [W]
    """
    # Extract Npt data
    pts = NPT_dict['pts'][0, :].reshape(NPT_dict['pts'].shape[1], 1)
    negpts = NPT_dict['negpts'][0, :].reshape(NPT_dict['negpts'].shape[1], 1)
    Imod_pts = NPT_dict['Imod_pts'][0, :].reshape(
        NPT_dict['Imod_pts'].shape[1], 1)
    Imod_negpts = NPT_dict['Imod_negpts'][0, :].reshape(
        NPT_dict['Imod_negpts'].shape[1], 1)
    Npts = NPT_dict['Npts']
    # iterate over substrings
    Isubstr, Vsubstr, Isc_substr, Imax_substr = [], [], [], []
    Isubstr_pre_bypass, Vsubstr_pre_bypass = [], []
    substr_bypass = []
    for substr_idx, substr in enumerate(cell_pos):
        # check if cells are in series or any crosstied circuits
        if all(not r['crosstie'] for c in substr for r in c):
            idxs = [r['idx'] for c in substr for r in c]
            # t0 = time.time()
            IatVrbd = np.asarray(
                [np.interp(vrbd, v, i) for vrbd, v, i in
                 zip(VRBD[idxs], Vcell[idxs], Icell[idxs])]
            )
            # print('Time elapsed to run loopy interp: ' + str(time.time() - t0) + ' s')
            # Icell_red = Icell[idxs]
            # Vcell_red = Vcell[idxs]
            # VRBD_red = VRBD[idxs].reshape(len(VRBD[idxs]),1)
            # t0 = time.time()
            # IatVrbd = interp2d_wrap(Vcell_red, VRBD_red, Icell_red).flatten()
            # print('Time elapsed to run vectorized interp: ' + str(time.time() - t0) + ' s')
            Isub, Vsub = calcSeries(
                Icell[idxs], Vcell[idxs], Isc[idxs].mean(),
                IatVrbd.max(), Imod_pts, Imod_negpts, Npts
            )
        elif all(r['crosstie'] for c in substr for r in c):
            Irows, Vrows = [], []
            Isc_rows, Imax_rows = [], []
            for row in zip(*substr):
                idxs = [c['idx'] for c in row]
                Irow, Vrow = calcParallel(
                    Icell[idxs], Vcell[idxs],
                    Voc[idxs].max(), VRBD.min(), negpts, pts, Npts
                )
                Irows.append(Irow)
                Vrows.append(Vrow)
                Isc_rows.append(np.interp(np.float64(0), Vrow, Irow))
                Imax_rows.append(Irow.max())
            Irows, Vrows = np.asarray(Irows), np.asarray(Vrows)
            Isc_rows = np.asarray(Isc_rows)
            Imax_rows = np.asarray(Imax_rows)
            Isub, Vsub = calcSeries(
                Irows, Vrows, Isc_rows.mean(), Imax_rows.max(),
                Imod_pts, Imod_negpts, Npts
            )
        else:
            IVall_cols = []
            prev_col = None
            IVprev_cols = []
            for col in substr:
                IVcols = []
                is_first = True
                # combine series between crossties
                for idxs in pvconstants.get_series_cells(col, prev_col):
                    if not idxs:
                        # first row should always be empty since it must be
                        # crosstied
                        is_first = False
                        continue
                    elif is_first:
                        raise Exception(
                            "First row and last rows must be crosstied."
                        )
                    elif len(idxs) > 1:
                        IatVrbd = np.asarray(
                            [np.interp(vrbd, v, i) for vrbd, v, i in
                             zip(VRBD[idxs], Vcell[idxs],
                                 Icell[idxs])]
                        )
                        Icol, Vcol = calcSeries(
                            Icell[idxs], Vcell[idxs],
                            Isc[idxs].mean(), IatVrbd.max(),
                            Imod_pts, Imod_negpts, Npts
                        )
                    else:
                        Icol, Vcol = Icell[idxs], Vcell[idxs]
                    IVcols.append([Icol, Vcol])
                # append IVcols and continue
                IVprev_cols.append(IVcols)
                if prev_col:
                    # if circuits are same in both columns then continue
                    if not all(icol['crosstie'] == jcol['crosstie']
                               for icol, jcol in zip(prev_col, col)):
                        # combine crosstied circuits
                        Iparallel, Vparallel = combine_parallel_circuits(
                            IVprev_cols, pvconstants,
                            negpts, pts, Imod_pts, Imod_negpts, Npts
                        )
                        IVall_cols.append([Iparallel, Vparallel])
                        # reset prev_col
                        prev_col = None
                        IVprev_cols = []
                        continue
                # set prev_col and continue
                prev_col = col
            # combine any remaining crosstied circuits in substring
            if not IVall_cols:
                # combine crosstied circuits
                Isub, Vsub = combine_parallel_circuits(
                    IVprev_cols, pvconstants,
                    negpts, pts, Imod_pts, Imod_negpts, Npts
                )
            else:
                Iparallel, Vparallel = zip(*IVall_cols)
                Iparallel = np.asarray(Iparallel)
                Vparallel = np.asarray(Vparallel)
                Isub, Vsub = calcParallel(
                    Iparallel, Vparallel, Vparallel.max(), Vparallel.min(),
                    negpts, pts, Npts
                )

        Isubstr_pre_bypass.append(Isub.copy())
        Vsubstr_pre_bypass.append(Vsub.copy())

        Vbypass_config = parse_diode_config(Vbypass, cell_pos)
        if Vbypass_config == DEFAULT_BYPASS:
            bypassed = Vsub < Vbypass
            Vsub[bypassed] = Vbypass
        elif Vbypass_config == CUSTOM_SUBSTR_BYPASS:
            if Vbypass[substr_idx] is None:
                # no bypass for this substring
                bypassed = np.zeros(Vsub.shape, dtype=bool)
                pass
            else:
                # bypass the substring
                bypassed = Vsub < Vbypass[substr_idx]
                Vsub[bypassed] = Vbypass[substr_idx]
        elif Vbypass_config == MODULE_BYPASS:
            # module bypass value will be assigned after the for loop for substrings is over
            bypassed = np.zeros(Vsub.shape, dtype=bool)
            pass

        Isubstr.append(Isub)
        Vsubstr.append(Vsub)
        Isc_substr.append(np.interp(np.float64(0), Vsub, Isub))
        Imax_substr.append(Isub.max())
        substr_bypass.append(bypassed)

    Isubstr, Vsubstr = np.asarray(Isubstr), np.asarray(Vsubstr)
    substr_bypass = np.asarray(substr_bypass)
    Isubstr_pre_bypass, Vsubstr_pre_bypass = np.asarray(Isubstr_pre_bypass), np.asarray(Vsubstr_pre_bypass)
    Isc_substr = np.asarray(Isc_substr)
    Imax_substr = np.asarray(Imax_substr)
    if outer == 'series':
        Imod, Vmod, bypassed_mod = calcSeries_with_bypass(
            Isubstr, Vsubstr, Isc_substr.mean(), Imax_substr.max(),
            Imod_pts, Imod_negpts, Npts, substr_bypass, run_bpact=run_bpact
        )
    else:
        Imod, Vmod, bypassed_mod = calcParallel_with_bypass(
            Isubstr, Vsubstr, Vsubstr.max(), Vsubstr.min(),
            Imod_negpts, Imod_pts, Npts, substr_bypass, run_bpact=run_bpact)

    # if entire module has only one bypass diode
    if Vbypass_config == MODULE_BYPASS:
        bypassed = Vmod < Vbypass[0]
        Vmod[bypassed] = Vbypass[0]
        bypassed_mod = bypassed[np.newaxis, ...]
    else:
        pass

    Pmod = Imod * Vmod
    return (Imod, Vmod, Pmod, Isubstr, Vsubstr, Isc.mean(), Isubstr_pre_bypass, Vsubstr_pre_bypass, bypassed_mod)



# -*- coding: utf-8 -*-

"""
This module contains the :class:`~pvmismatch.pvmismatch_lib.pvcell.PVcell`
object which is used by modules, strings and systems.
"""

from __future__ import absolute_import
from future.utils import iteritems
from .pvconstants import PVconstants
import numpy as np
from matplotlib import pyplot as plt
from scipy.optimize import newton

# Defaults
RS = 0.004267236774264931  # [ohm] series resistance
RSH = 10.01226369025448  # [ohm] shunt resistance
ISAT1_T0 = 2.286188161253440E-11  # [A] diode one saturation current
ISAT2_T0 = 1.117455042372326E-6  # [A] diode two saturation current
ISC0_T0 = 6.3056  # [A] reference short circuit current
TCELL = 298.15  # [K] cell temperature
ARBD = 1.036748445065697E-4  # reverse breakdown coefficient 1
BRBD = 0.  # reverse breakdown coefficient 2
VRBD_ = -5.527260068445654  # [V] reverse breakdown voltage
NRBD = 3.284628553041425  # reverse breakdown exponent
EG = 1.1  # [eV] band gap of cSi
ALPHA_ISC = 0.0003551  # [1/K] short circuit current temperature coefficient
EPS = np.finfo(np.float64).eps


class PVcell(object):
    """
    Class for PV cells.

    :param Rs: series resistance [ohms]
    :param Rsh: shunt resistance [ohms]
    :param Isat1_T0: first saturation diode current at ref temp [A]
    :param Isat2_T0: second saturation diode current [A]
    :param Isc0_T0: short circuit current at ref temp [A]
    :param aRBD: reverse breakdown coefficient 1
    :param bRBD: reverse breakdown coefficient 2
    :param VRBD: reverse breakdown voltage [V]
    :param nRBD: reverse breakdown exponent
    :param Eg: band gap [eV]
    :param alpha_Isc: short circuit current temp coeff [1/K]
    :param Tcell: cell temperature [K]
    :param Ee: incident effective irradiance [suns]
    :param pvconst: configuration constants object
    :type pvconst: :class:`~pvmismatch.pvmismatch_lib.pvconstants.PVconstants`
    """

    _calc_now = False  #: if True ``calcCells()`` is called in ``__setattr__``

    def __init__(self, Rs=RS, Rsh=RSH, Isat1_T0=ISAT1_T0, Isat2_T0=ISAT2_T0,
                 Isc0_T0=ISC0_T0, aRBD=ARBD, bRBD=BRBD, VRBD=VRBD_,
                 nRBD=NRBD, Eg=EG, alpha_Isc=ALPHA_ISC,
                 Tcell=TCELL, Ee=1., pvconst=PVconstants()):
        # user inputs
        self.Rs = Rs  #: [ohm] series resistance
        self.Rsh = Rsh  #: [ohm] shunt resistance
        self.Isat1_T0 = Isat1_T0  #: [A] diode one sat. current at T0
        self.Isat2_T0 = Isat2_T0  #: [A] diode two saturation current
        self.Isc0_T0 = Isc0_T0  #: [A] short circuit current at T0
        self.aRBD = aRBD  #: reverse breakdown coefficient 1
        self.bRBD = bRBD  #: reverse breakdown coefficient 2
        self.VRBD = VRBD  #: [V] reverse breakdown voltage
        self.nRBD = nRBD  #: reverse breakdown exponent
        self.Eg = Eg  #: [eV] band gap of cSi
        self.alpha_Isc = alpha_Isc  #: [1/K] short circuit temp. coeff.
        self.Tcell = Tcell  #: [K] cell temperature
        self.Ee = Ee  #: [suns] incident effective irradiance on cell
        self.pvconst = pvconst  #: configuration constants
        self.Icell = None  #: cell currents on IV curve [A]
        self.Vcell = None  #: cell voltages on IV curve [V]
        self.Pcell = None  #: cell power on IV curve [W]
        self.VocSTC = self._VocSTC()  #: estimated Voc at STC [V]
        # set calculation flag
        self._calc_now = True  # overwrites the class attribute

    def __str__(self):
        fmt = '<PVcell(Ee=%g[suns], Tcell=%g[K], Isc=%g[A], Voc=%g[V])>'
        return fmt % (self.Ee, self.Tcell, self.Isc, self.Voc)

    def __repr__(self):
        return str(self)

    def __setattr__(self, key, value):
        # check for floats
        try:
            value = np.float64(value)
        except (TypeError, ValueError):
            pass  # fail silently if not float, eg: pvconst or _calc_now
        super(PVcell, self).__setattr__(key, value)
        # recalculate IV curve
        if self._calc_now:
            Icell, Vcell, Pcell = self.calcCell()
            self.__dict__.update(Icell=Icell, Vcell=Vcell, Pcell=Pcell)

    def update(self, **kwargs):
        """
        Update user-defined constants.
        """
        # turn off calculation flag until all attributes are updated
        self._calc_now = False
        # don't use __dict__.update() instead use setattr() to go through
        # custom __setattr__() so that numbers are cast to floats
        for k, v in iteritems(kwargs):
            setattr(self, k, v)
        self._calc_now = True  # recalculate

    @property
    def Vt(self):
        """
        Thermal voltage in volts.
        """
        return self.pvconst.k * self.Tcell / self.pvconst.q

    @property
    def Isc(self):
        return self.Ee * self.Isc0

    @property
    def Aph(self):
        """
        Photogenerated current coefficient, non-dimensional.
        """
        # Aph is undefined (0/0) if there is no irradiance
        if self.Isc == 0:
            return np.nan
        # short current (SC) conditions (Vcell = 0)
        Vdiode_sc = self.Isc * self.Rs  # diode voltage at SC
        Idiode1_sc = self.Isat1 * (np.exp(Vdiode_sc / self.Vt) - 1.)
        Idiode2_sc = self.Isat2 * (np.exp(Vdiode_sc / 2. / self.Vt) - 1.)
        Ishunt_sc = Vdiode_sc / self.Rsh  # diode voltage at SC
        # photogenerated current coefficient
        return 1. + (Idiode1_sc + Idiode2_sc + Ishunt_sc) / self.Isc

    @property
    def Isat1(self):
        """
        Diode one saturation current at Tcell in amps.
        """
        _Tstar = self.Tcell ** 3. / self.pvconst.T0 ** 3.  # scaled temperature
        _inv_delta_T = 1. / self.pvconst.T0 - 1. / self.Tcell  # [1/K]
        _expTstar = np.exp(
            self.Eg * self.pvconst.q / self.pvconst.k * _inv_delta_T
        )
        return self.Isat1_T0 * _Tstar * _expTstar  # [A] Isat1(Tcell)

    @property
    def Isat2(self):
        """
        Diode two saturation current at Tcell in amps.
        """
        _Tstar = self.Tcell ** 3. / self.pvconst.T0 ** 3.  # scaled temperature
        _inv_delta_T = 1. / self.pvconst.T0 - 1. / self.Tcell  # [1/K]
        _expTstar = np.exp(
            self.Eg * self.pvconst.q / (2.0 * self.pvconst.k) * _inv_delta_T
        )
        return self.Isat2_T0 * _Tstar * _expTstar  # [A] Isat2(Tcell)

    @property
    def Isc0(self):
        """
        Short circuit current at Tcell in amps.
        """
        _delta_T = self.Tcell - self.pvconst.T0  # [K] temperature difference
        return self.Isc0_T0 * (1. + self.alpha_Isc * _delta_T)  # [A] Isc0

    @property
    def Voc(self):
        """
        Estimate open circuit voltage of cells.
        Returns Voc : numpy.ndarray of float, estimated open circuit voltage
        """
        C = self.Aph * self.Isc + self.Isat1 + self.Isat2
        delta = self.Isat2 ** 2. + 4. * self.Isat1 * C
        return self.Vt * np.log(
            ((-self.Isat2 + np.sqrt(delta)) / 2. / self.Isat1) ** 2.
        )

    def _VocSTC(self):
        """
        Estimate open circuit voltage of cells.
        Returns Voc : numpy.ndarray of float, estimated open circuit voltage
        """
        Vdiode_sc = self.Isc0_T0 * self.Rs  # diode voltage at SC
        Idiode1_sc = self.Isat1_T0 * (np.exp(Vdiode_sc / self.Vt) - 1.)
        Idiode2_sc = self.Isat2_T0 * (np.exp(Vdiode_sc / 2. / self.Vt) - 1.)
        Ishunt_sc = Vdiode_sc / self.Rsh  # diode voltage at SC
        # photogenerated current coefficient
        Aph = 1. + (Idiode1_sc + Idiode2_sc + Ishunt_sc) / self.Isc0_T0
        # estimated Voc at STC
        C = Aph * self.Isc0_T0 + self.Isat1_T0 + self.Isat2_T0
        delta = self.Isat2_T0 ** 2. + 4. * self.Isat1_T0 * C
        return self.Vt * np.log(
            ((-self.Isat2_T0 + np.sqrt(delta)) / 2. / self.Isat1_T0) ** 2.
        )

    @property
    def Igen(self):
        """
        Photovoltaic generated light current (AKA IL or Iph)
        Returns Igen : numpy.ndarray of float, PV generated light current [A]

        Photovoltaic generated light current is zero if irradiance is zero.
        """
        if self.Ee == 0:
            return 0
        return self.Aph * self.Isc

    def calcCell(self):
        """
        Calculate cell I-V curves.
        Returns (Icell, Vcell, Pcell) : tuple of numpy.ndarray of float
        """
        Vreverse = self.VRBD * self.pvconst.negpts
        Vff = self.Voc
        delta_Voc = self.VocSTC - self.Voc
        # to make sure that the max voltage is always in the 4th quadrant, add
        # a third set of points log spaced with decreasing density, from Voc to
        # Voc @ STC unless Voc *is* Voc @ STC, then use an arbitrary voltage at
        # 80% of Voc as an estimate of Vmp assuming a fill factor of 80% and
        # Isc close to Imp, or if Voc > Voc @ STC, then use Voc as the max
        if delta_Voc == 0:
            Vff = 0.8 * self.Voc
            delta_Voc = 0.2 * self.Voc
        elif delta_Voc < 0:
            Vff = self.VocSTC
            delta_Voc = -delta_Voc
        Vquad4 = Vff + delta_Voc * self.pvconst.Vmod_q4pts
        Vforward = Vff * self.pvconst.pts
        Vdiode = np.concatenate((Vreverse, Vforward, Vquad4), axis=0)
        Idiode1 = self.Isat1 * (np.exp(Vdiode / self.Vt) - 1.)
        Idiode2 = self.Isat2 * (np.exp(Vdiode / 2. / self.Vt) - 1.)
        Ishunt = Vdiode / self.Rsh
        fRBD = 1. - Vdiode / self.VRBD
        # use epsilon = 2.2204460492503131e-16 to avoid "divide by zero"
        fRBD[fRBD == 0] = EPS
        Vdiode_norm = Vdiode / self.Rsh / self.Isc0_T0
        fRBD = self.Isc0_T0 * fRBD ** (-self.nRBD)
        IRBD = (self.aRBD * Vdiode_norm + self.bRBD * Vdiode_norm ** 2) * fRBD
        Icell = self.Igen - Idiode1 - Idiode2 - Ishunt - IRBD
        Vcell = Vdiode - Icell * self.Rs
        Pcell = Icell * Vcell
        return Icell, Vcell, Pcell

    # diode model
    #  *-->--*--->---*--Rs->-Icell--+
    #  ^     |       |              ^
    #  |     |       |              |
    # Igen  Idiode  Ishunt         Vcell
    #  |     |       |              |
    #  |     v       v              v
    #  *--<--*---<---*--<-----------=
    # http://en.wikipedia.org/wiki/Diode_modelling#Shockley_diode_model
    # http://en.wikipedia.org/wiki/Diode#Shockley_diode_equation
    # http://en.wikipedia.org/wiki/William_Shockley

    @staticmethod
    def f_Icell(Icell, Vcell, Igen, Rs, Vt, Isat1, Isat2, Rsh):
        """
        Objective function for Icell.
        :param Icell: cell current [A]
        :param Vcell: cell voltage [V]
        :param Igen: photogenerated current at Tcell and Ee [A]
        :param Rs: series resistance [ohms]
        :param Vt: thermal voltage [V]
        :param Isat1: first diode saturation current at Tcell [A]
        :param Isat2: second diode saturation current [A]
        :param Rsh: shunt resistance [ohms]
        :return: residual = (Icell - Icell0) [A]
        """
        # arbitrary current condition
        Vdiode = Vcell + Icell * Rs  # diode voltage
        Idiode1 = Isat1 * (np.exp(Vdiode / Vt) - 1.)  # diode current
        Idiode2 = Isat2 * (np.exp(Vdiode / 2. / Vt) - 1.)  # diode current
        Ishunt = Vdiode / Rsh  # shunt current
        return Igen - Idiode1 - Idiode2 - Ishunt - Icell

    def calcIcell(self, Vcell):
        """
        Calculate Icell as a function of Vcell.
        :param Vcell: cell voltage [V]
        :return: Icell
        """
        args = (np.float64(Vcell), self.Igen, self.Rs, self.Vt,
                self.Isat1, self.Isat2, self.Rsh)
        return newton(self.f_Icell, x0=self.Isc, args=args)

    @staticmethod
    def f_Vcell(Vcell, Icell, Igen, Rs, Vt, Isat1, Isat2, Rsh):
        return PVcell.f_Icell(Icell, Vcell, Igen, Rs, Vt, Isat1, Isat2, Rsh)

    def calcVcell(self, Icell):
        """
        Calculate Vcell as a function of Icell.
        :param Icell: cell current [A]
        :return: Vcell
        """
        args = (np.float64(Icell), self.Igen, self.Rs, self.Vt,
                self.Isat1, self.Isat2, self.Rsh)
        return newton(self.f_Vcell, x0=self.Voc, args=args)

    def plot(self):
        """
        Plot cell I-V curve.
        Returns cellPlot : matplotlib.pyplot figure
        """
        cell_plot = plt.figure()
        plt.subplot(2, 2, 1)
        plt.plot(self.Vcell, self.Icell)
        plt.title('Cell Reverse I-V Characteristics')
        plt.ylabel('Cell Current, I [A]')
        plt.xlim(self.VRBD - 1, 0)
        plt.ylim(0, self.Isc + 10)
        plt.grid()
        plt.subplot(2, 2, 2)
        plt.plot(self.Vcell, self.Icell)
        plt.title('Cell Forward I-V Characteristics')
        plt.ylabel('Cell Current, I [A]')
        plt.xlim(0, self.Voc)
        plt.ylim(0, self.Isc + 1)
        plt.grid()
        plt.subplot(2, 2, 3)
        plt.plot(self.Vcell, self.Pcell)
        plt.title('Cell Reverse P-V Characteristics')
        plt.xlabel('Cell Voltage, V [V]')
        plt.ylabel('Cell Power, P [W]')
        plt.xlim(self.VRBD - 1, 0)
        plt.ylim((self.Isc + 10) * (self.VRBD - 1), -1)
        plt.grid()
        plt.subplot(2, 2, 4)
        plt.plot(self.Vcell, self.Pcell)
        plt.title('Cell Forward P-V Characteristics')
        plt.xlabel('Cell Voltage, V [V]')
        plt.ylabel('Cell Power, P [W]')
        plt.xlim(0, self.Voc)
        plt.ylim(0, (self.Isc + 1) * self.Voc)
        plt.grid()
        plt.tight_layout()
        return cell_plot



try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    # for python < 3.8
    from importlib_metadata import PackageNotFoundError, version

try:
    __version__ = version(__package__)
except PackageNotFoundError:
    __version__ = "0+unknown"



# -*- coding: utf-8 -*-
"""Vectorized pvstring."""

import numpy as np
from .utils import isin_nd
from .circuit_comb import calcSeries_with_bypass


def calcStrings(Ee_str, Ee_mod, mod_data, NPT_dict, run_bpact=True, run_annual=False):
    """
    Generate all string IV curves and store results in a dictionary.

    Parameters
    ----------
    Ee_str : numpy.ndarray
        4-D array containing the Irradiance at the cell level for all modules in each string.
    Ee_mod : numpy.ndarray
        3-D array containing the Irradiance at the cell level for all modules.
    mod_data : dict
        Dictionary containing module IV curves.
    NPT_dict : numpy.ndarray
        NPTs dictionary from the cell data dictionary generated by pvcell.
    run_bpact : bool, optional
        Flag to run bypass diode activation logic. The default is True.
    run_annual : bool, optional
        Flag to delete large bypass diode activation array in case of an annual simulation. The default is False.

    Returns
    -------
    str_data : dict
        Dictionary containing string IV curves.

    """
    I_str_curves = []
    V_str_curves = []
    P_str_curves = []
    Bypass_str_curves = []
    for idx_str in range(Ee_str.shape[0]):

        # 1 String
        Ee_str1 = Ee_str[idx_str]
        # Extract mod IV curves
        str_in_mod = isin_nd(Ee_mod, Ee_str1)
        Imod_red = mod_data['Imod'][str_in_mod, :]
        Vmod_red = mod_data['Vmod'][str_in_mod, :]
        meanIsc_red = mod_data['mean_Isc'][str_in_mod]
        if run_bpact:
            Bypassed_substr_red = mod_data['Bypassed_substr'][str_in_mod, :, :]
        else:
            Bypassed_substr_red = np.nan
        u, inverse, counts = np.unique(Ee_str1, axis=0, return_inverse=True,
                                       return_counts=True)
        # Expand for Str curves
        Imod = Imod_red[inverse, :]
        Vmod = Vmod_red[inverse, :]
        meanIsc = meanIsc_red[inverse]
        if run_bpact:
            Bypassed_substr = Bypassed_substr_red[inverse, :, :]
        else:
            Bypassed_substr = np.nan
        Imod_pts = NPT_dict['Imod_pts'][0, :].reshape(
            NPT_dict['Imod_pts'].shape[1], 1)
        Imod_negpts = NPT_dict['Imod_negpts'][0, :].reshape(
            NPT_dict['Imod_negpts'].shape[1], 1)
        Npts = NPT_dict['Npts']
        # Run String Circuit model
        Istring, Vstring, bypassed_str = calcSeries_with_bypass(
            Imod, Vmod, meanIsc.mean(), Imod.max(), Imod_pts, Imod_negpts, Npts, Bypassed_substr, run_bpact=run_bpact)
        Pstring = Istring * Vstring
        # If it is an annual simulation, delete the bypassed related variables
        if run_annual:
            del Bypassed_substr_red
            del Bypassed_substr

        I_str_curves.append(np.reshape(Istring, (1, len(Istring))))
        V_str_curves.append(np.reshape(Vstring, (1, len(Vstring))))
        P_str_curves.append(np.reshape(Pstring, (1, len(Pstring))))
        if run_bpact:
            Bypass_str_curves.append(np.reshape(bypassed_str, (1,
                                                               bypassed_str.shape[0],
                                                               bypassed_str.shape[1],
                                                               bypassed_str.shape[2])))
        else:
            Bypass_str_curves.append(bypassed_str)

    I_str_curves = np.concatenate(I_str_curves, axis=0)
    V_str_curves = np.concatenate(V_str_curves, axis=0)
    P_str_curves = np.concatenate(P_str_curves, axis=0)
    if run_bpact:
        Bypass_str_curves = np.concatenate(Bypass_str_curves, axis=0)
    else:
        Bypass_str_curves = np.array(Bypass_str_curves)
    # If it is an annual simulation, delete the bypassed related variables
    if run_annual:
        del mod_data['Bypassed_substr']

    # Store results in a dict
    str_data = dict()
    str_data['Istring'] = I_str_curves
    str_data['Vstring'] = V_str_curves
    str_data['Pstring'] = P_str_curves
    str_data['Bypassed_substr'] = Bypass_str_curves

    return str_data



# -*- coding: utf-8 -*-
"""Circuit combination and bypass diode configuration functions that are vectorized."""

import numpy as np
from scipy.interpolate import interp1d
from .pvmismatch import pvconstants

# ------------------------------------------------------------------------------
# CIRCUIT COMBINATION FUNCTIONS-------------------------------------------------

DEFAULT_BYPASS = 0
MODULE_BYPASS = 1
CUSTOM_SUBSTR_BYPASS = 2


def calcSeries(Curr, V, meanIsc, Imax, Imod_pts, Imod_negpts, npts):
    """
    Calculate IV curve for cells and substrings in series.

    Given current and voltage in increasing order by voltage,
    the average short circuit current and the max current at the breakdown voltage.

    :param Curr: cell or substring currents [A]
    :param V: cell or substring voltages [V]
    :param meanIsc: average short circuit current [A]
    :param Imax: maximum current [A]
    :return: current [A] and voltage [V] of series

    """
    # make sure all inputs are numpy arrays, but don't make extra copies
    Curr = np.asarray(Curr)  # currents [A]
    V = np.asarray(V)  # voltages [V]
    meanIsc = np.asarray(meanIsc)  # mean Isc [A]
    Imax = np.asarray(Imax)  # max current [A]
    # create array of currents optimally spaced from mean Isc to  max VRBD
    Ireverse = (Imax - meanIsc) * Imod_pts + meanIsc
    Imin = np.minimum(Curr.min(), 0.)  # minimum cell current, at most zero
    # range of currents in forward bias from min current to mean Isc
    Iforward = (Imin - meanIsc) * Imod_negpts + meanIsc
    # create range for interpolation from forward to reverse bias
    Itot = np.concatenate((Iforward, Ireverse), axis=0).flatten()
    Vtot = np.zeros((2 * npts,))
    # add up all series cell voltages
    # t0 = time.time()
    for i, v in zip(Curr, V):
        # interp requires x, y to be sorted by x in increasing order
        Vtot += pvconstants.npinterpx(Itot, np.flipud(i), np.flipud(v))
    # print('Time elapsed to run loopy interp: ' + str(time.time() - t0) + ' s')
    # Itot = np.repeat(Itot[:,np.newaxis], Curr.shape[0], axis=1).T
    # t0 = time.time()
    # Vtot = interp2d_wrap(np.flip(Curr, axis = 1), Itot, np.flip(V, axis = 1))
    # Vtot = Vtot.sum(axis=1)
    # print('Time elapsed to run vectorized interp: ' + str(time.time() - t0) + ' s')
    return np.flipud(Itot), np.flipud(Vtot)


def calcSeries_with_bypass(Curr, V, meanIsc, Imax, Imod_pts, Imod_negpts, npts, substr_bypass, run_bpact=True):
    """
    Calculate IV curve for cells and substrings in series.

    Given current and voltage in increasing order by voltage,
    the average short circuit current and the max current at
    the breakdown voltage.

    :param Curr: cell or substring currents [A]
    :param V: cell or substring voltages [V]
    :param meanIsc: average short circuit current [A]
    :param Imax: maximum current [A]
    :return: current [A] and voltage [V] of series
    """
    # make sure all inputs are numpy arrays, but don't make extra copies
    Curr = np.asarray(Curr)  # currents [A]
    V = np.asarray(V)  # voltages [V]
    meanIsc = np.asarray(meanIsc)  # mean Isc [A]
    Imax = np.asarray(Imax)  # max current [A]

    # create array of currents optimally spaced from mean Isc to  max VRBD
    Ireverse = (Imax - meanIsc) * Imod_pts + meanIsc
    Imin = np.minimum(Curr.min(), 0.)  # minimum cell current, at most zero
    # range of currents in forward bias from min current to mean Isc
    Iforward = (Imin - meanIsc) * Imod_negpts + meanIsc
    # create range for interpolation from forward to reverse bias
    Itot = np.concatenate((Iforward, Ireverse), axis=0).flatten()
    Vtot = np.zeros((2 * npts,))

    # t0 = time.time()
    # add up all series cell voltages
    for i, v in zip(Curr, V):
        # interp requires x, y to be sorted by x in increasing order
        Vtot += pvconstants.npinterpx(Itot, np.flipud(i), np.flipud(v))

    # Logic to check the shape of substr_bypass and ensure interpolations is done correctly
    # Case 1 (Module): I, V, sb have same shape (n x npt)
    # Case 2 (String): I, & V and sb have different dimensions (n x npt and m x n x npt)
    if run_bpact:
        if len(Curr.shape) == len(substr_bypass.shape):
            bypassed_mod = []
            for i, v, bypassed in zip(Curr, V, substr_bypass):
                if ~bypassed.all():
                    bp_interp = bypassed[0]*np.ones(Itot.shape)
                else:
                    interpolator = interp1d(np.flipud(i), np.flipud(bypassed),
                                            kind='previous', fill_value='extrapolate')
                    bp_interp = interpolator(Itot)
                bp_interp = bp_interp.astype(bool)
                bypassed_mod.append(np.flipud(bp_interp))
            bypassed_mod = np.asarray(bypassed_mod)
        else:
            bypassed_mod = np.zeros((Curr.shape[0], substr_bypass.shape[1], 2 * npts))
            idx_mod = 0
            for i, v, bypassed_strs in zip(Curr, V, substr_bypass):
                for idx_substr in range(bypassed_strs.shape[0]):
                    bypassed = bypassed_strs[idx_substr, :]
                    if ~bypassed.all() or bypassed.all():
                        bp_interp = bypassed[0]*np.ones(Itot.shape)
                    else:
                        interpolator = interp1d(np.flipud(i), np.flipud(bypassed),
                                                kind='previous', fill_value='extrapolate')
                        bp_interp = interpolator(Itot)
                    bp_interp = bp_interp.astype(bool)
                    bypassed_mod[idx_mod, idx_substr, :] = np.flipud(bp_interp)
                idx_mod += 1
    else:
        bypassed_mod = np.nan
    return np.flipud(Itot), np.flipud(Vtot), bypassed_mod


def calcParallel(Curr, V, Vmax, Vmin, negpts, pts, npts):
    """
    Calculate IV curve for cells and substrings in parallel.

    :param Curr: currents [A]
    :type: Curr: list, :class:`numpy.ndarray`
    :param V: voltages [V]
    :type: V: list, :class:`numpy.ndarray`
    :param Vmax: max voltage limit, should be max Voc [V]
    :param Vmin: min voltage limit, could be zero or Vrbd [V]
    """
    Curr, V = np.asarray(Curr), np.asarray(V)
    Vmax = np.asarray(Vmax)
    Vmin = np.asarray(Vmin)
    Vreverse = Vmin * negpts
    Vforward = Vmax * pts
    Vtot = np.concatenate((Vreverse, Vforward), axis=0).flatten()
    Itot = np.zeros((2 * npts,))
    for i, v in zip(Curr, V):
        Itot += pvconstants.npinterpx(Vtot, v, i)
    return Itot, Vtot


def calcParallel_with_bypass(Curr, V, Vmax, Vmin, negpts, pts, npts, substr_bypass, run_bpact=True):
    """
    Calculate IV curve for cells and substrings in parallel.

    :param Curr: currents [A]
    :type: Curr: list, :class:`numpy.ndarray`
    :param V: voltages [V]
    :type: V: list, :class:`numpy.ndarray`
    :param Vmax: max voltage limit, should be max Voc [V]
    :param Vmin: min voltage limit, could be zero or Vrbd [V]
    """
    Curr, V = np.asarray(Curr), np.asarray(V)
    Vmax = np.asarray(Vmax)
    Vmin = np.asarray(Vmin)
    Vreverse = Vmin * negpts
    Vforward = Vmax * pts
    Vtot = np.concatenate((Vreverse, Vforward), axis=0).flatten()
    Itot = np.zeros((2 * npts,))
    for i, v in zip(Curr, V):
        Itot += pvconstants.npinterpx(Vtot, v, i)

    if run_bpact:
        if len(Curr.shape) == len(substr_bypass.shape):
            bypassed_mod = []
            for i, v, bypassed in zip(Curr, V, substr_bypass):
                if ~bypassed.all():
                    bp_interp = bypassed[0]*np.ones(Vtot.shape)
                else:
                    interpolator = interp1d(v, bypassed, kind='previous',
                                            fill_value='extrapolate')
                    bp_interp = interpolator(Vtot)
                bp_interp = bp_interp.astype(bool)
                bypassed_mod.append(bp_interp)
            bypassed_mod = np.asarray(bypassed_mod)
            # bypassed_mod = interp2d_wrap(V, Vtot1, substr_bypass.astype(float), kind='previous')
            # bypassed_mod = np.round(bypassed_mod).astype(bool)
        else:
            bypassed_mod = np.zeros((Curr.shape[0], substr_bypass.shape[1],
                                    substr_bypass.shape[2], 2 * npts), dtype=bool)
            idx_str = 0
            for i, v, bypassed_strs in zip(Curr, V, substr_bypass):
                idx_mod = 0
                for idx_mod in range(bypassed_strs.shape[0]):
                    for idx_substr in range(bypassed_strs.shape[1]):
                        bypassed = bypassed_strs[idx_mod, idx_substr, :]
                        if ~bypassed.all() or bypassed.all():
                            bp_interp = bypassed[0]*np.ones(Vtot.shape)
                        else:
                            interpolator = interp1d(v, bypassed, kind='previous',
                                                    fill_value='extrapolate')
                            bp_interp = interpolator(Vtot)
                        bp_interp = bp_interp.astype(bool)
                        bypassed_mod[idx_str, idx_mod, idx_substr, :] = bp_interp
                    idx_mod += 1
                idx_str += 1
    else:
        bypassed_mod = np.nan

    return Itot, Vtot, bypassed_mod


def combine_parallel_circuits(IVprev_cols, pvconst,
                              negpts, pts, Imod_pts, Imod_negpts, npts):
    """
    Combine crosstied circuits in a substring.

    :param IVprev_cols: lists of IV curves of crosstied and series circuits
    :return:
    """
    # combine crosstied circuits
    Irows, Vrows = [], []
    Isc_rows, Imax_rows = [], []
    for IVcols in zip(*IVprev_cols):
        Iparallel, Vparallel = zip(*IVcols)
        Iparallel = np.asarray(Iparallel)
        Vparallel = np.asarray(Vparallel)
        Irow, Vrow = calcParallel(
            Iparallel, Vparallel, Vparallel.max(),
            Vparallel.min(), negpts, pts, npts
        )
        Irows.append(Irow)
        Vrows.append(Vrow)
        Isc_rows.append(np.interp(np.float64(0), Vrow, Irow))
        Imax_rows.append(Irow.max())
    Irows, Vrows = np.asarray(Irows), np.asarray(Vrows)
    Isc_rows = np.asarray(Isc_rows)
    Imax_rows = np.asarray(Imax_rows)
    return calcSeries(
        Irows, Vrows, Isc_rows.mean(), Imax_rows.max(),
        Imod_pts, Imod_negpts, npts
    )


def parse_diode_config(Vbypass, cell_pos):
    """
    Parse diode configuration from the Vbypass argument.

    :param Vbypass: Vbypass config
    :type Vbypass: float|list|tuple
    :param cell_pos:
    :type cell_pos:
    :return: bypass config
    :rtype: str
    """
    try:
        # check if float or list/tuple
        num_bypass = len(Vbypass)
    except TypeError:
        # float passed - default case - Vbypass across every cell string
        return DEFAULT_BYPASS
    else:
        # if only one value is passed in the list- assume only one
        # bypass diode  across the PV module
        if len(Vbypass) == 1:
            return MODULE_BYPASS
        # if more than 1 values are passed, apply them across
        # the cell strings in ascending order
        elif len(cell_pos) == num_bypass:
            return CUSTOM_SUBSTR_BYPASS
        else:
            print("wrong number of bypass diode values passed : %d" %
                  (len(Vbypass)))
